#include <u.h>
#include <libc.h>

#include "vga.h"

/*
 * Chips and Technologies 65550 (HiQV32).
 */
enum {
	Frx		= 0x3D0,	/* Flat Panel Extensions Index */
	Mrx		= 0x3D2,	/* Multimedia Extensions Index */
	Xrx		= 0x3D6,	/* Configuration Extensions Index */
};

typedef struct {
	uchar	fr[256];
	uchar	mr[256];
	uchar	xr[256];

	long	biosclk;
} Ct65550;

static uchar
ct65550xi(long port, uchar index)
{
	uchar data;

	outportb(port, index);
	data = inportb(port+1);

	return data;
}

static void
ct65550xo(long port, uchar index, uchar data)
{
	outportb(port, index);
	outportb(port+1, data);
}

static void
snarf(Vga* vga, Ctlr* ctlr)
{
	long i;
	Ct65550 *ct65550;

	if(vga->private == nil)
		vga->private = alloc(sizeof(Ct65550));
	ct65550 = vga->private;

	for(i = 0; i < 0x50; i++)
		ct65550->fr[i] = ct65550xi(Frx, i);
	for(i = 0; i < 0x60; i++)
		ct65550->mr[i] = ct65550xi(Mrx, i);
	for(i = 0x30; i < 0x80; i++)
		vga->crt[i] = vgaxi(Crtx, i);
	for(i = 0; i < 0x100; i++)
		ct65550->xr[i] = ct65550xi(Xrx, i);

	switch((ct65550->xr[0x43]>>1) & 0x03){
	case 0:
		vga->vmz = 1*1024*1024;
		break;

	case 1:
		vga->vmz = 2*1024*1024;
		break;

	default:
		vga->vmz = 16*1024*1024;
		break;
	}

	ctlr->flag |= Fsnarf;
}

static void
options(Vga *vganotused, Ctlr* ctlr)
{
	ctlr->flag |= Hlinear|Foptions;
}

void
clock(Vga* vga, Ctlr* ctlr)
{
	double f, fmin, fvco, m;
	ulong n, nmin, nmax, pd, rd;

	/*
	 * Constraints:
	 *  1.	1MHz <= Fref <= 60MHz
	 *  2.	150KHz <= Fref/(RD*N) <= 2MHz
	 *  3.	48MHz < Fvco <= 220MHz
	 *  4.	3 <= M <= 1023
	 *  5.	3 <= N <= 1023
	 */
	f = vga->f[0];

	/*
	 * Although pd = 0 is valid, things seems more stable
	 * starting at 1.
	 */
	for(pd = 1; pd < 6; pd++){
		f = vga->f[0]*(1<<pd);
		if(f > 48000000 && f <= 220000000)
			break;
	}
	if(pd >= 6)
		error("%s: dclk %lu out of range\n", ctlr->name, vga->f[0]);
	vga->p[0] = pd;

	/*
	 * The reference divisor has only two possible values, 1 and 4.
	 */
	fmin = f;
	for(rd = 1; rd <= 4; rd += 3){
		/*
		 * Find the range for n (constraint 2).
		 */
		for(nmin = 3; nmin <= 1023; nmin++){
			if(RefFreq/(rd*nmin) <= 2000000)
				break;
		}
		for(nmax = 1023; nmax >= 3; nmax--){
			if(RefFreq/(rd*nmax) >= 150000)
				break;
		}

		/*
		 * Now find the closest match for M and N.
		 */
		for(n = nmin; n < nmax; n++){
			for(m = 3; m <= vga->m[1]; m++){
				fvco = (RefFreq*4*m)/(rd*n);
				if(fvco < 48000000 || fvco > 220000000)
					continue;
				fvco -= f;
				if(fvco < 0)
					fvco = -fvco;
				if(fvco < fmin){
					fmin = fvco;
					vga->m[0] = m;
					vga->n[0] = n;
					vga->p[0] = pd;
					vga->r[0] = rd;
				}
			}
		}
	}

}

static void
biosclkinit(Vga* vga, Ctlr* ctlr)
{
	Ct65550 *ct65550;

	ct65550 = vga->private;
	ct65550->biosclk = 1;

	ctlr->flag |= Finit;
}

static void
init(Vga* vga, Ctlr* ctlr)
{
	ulong dclk;
	Ct65550 *ct65550;

	ct65550 = vga->private;

	if(vga->f[0] == 0)
		vga->f[0] = vga->mode->frequency;
	vga->misc &= ~0x0C;
	ct65550->fr[0x03] &= ~0x0C;

	if(ct65550->biosclk){
		vga->misc |= 0x08;
		ct65550->fr[0x03] |= 0x08;
	}
	else{
		/*
		 * Clock bits. If the desired video clock is
		 * one of the two standard VGA clocks it can just be
		 * set using bits <3:2> of vga->misc or Fr03,
		 * otherwise the DCLK PLL needs to be programmed.
		 */
		if(vga->f[0] == VgaFreq0)
			;
		else if(vga->f[0] == VgaFreq1){
			vga->misc |= 0x04;
			ct65550->fr[0x03] |= 0x04;
		}
		else{
			/*
			 * Check VCC to determine max clock.
			 * 5V allows a higher rate.
			 */
			if(ct65550->fr[0x0A] & 0x02)
				dclk = 110000000;
			else
				dclk = 80000000;
			if(vga->f[0] > dclk)
				error("%s: invalid dclk - %lu\n", ctlr->name, vga->f[0]);
	
			if(!ct65550->biosclk){
				vga->f[1] = dclk;
				vga->m[1] = 1023;
				clock(vga, ctlr);
				ct65550->xr[0xC8] = (vga->m[0]-2) & 0xFF;
				ct65550->xr[0xCA] = ((vga->m[0]-2) & 0x300)>>8;
				ct65550->xr[0xC9] = (vga->n[0]-2) & 0xFF;
				ct65550->xr[0xCA] |= ((vga->n[0]-2) & 0x300)>>4;
				ct65550->xr[0xCB] = (vga->p[0]<<4)|(vga->r[0] == 1);
			}
			vga->misc |= 0x08;
			ct65550->fr[0x03] |= 0x08;
		}
	}

	if(vga->mode->y >= 480)
		vga->misc &= ~0xC0;

	/*
	 * 10 bits should be enough, but set the extended mode anyway.
	 */
	ct65550->xr[0x09] |= 0x01;
	vga->crt[0x30] = ((vga->mode->vt-2)>>8) & 0x0F;
	vga->crt[0x31] = ((vga->mode->y-1)>>8) & 0x0F;
	vga->crt[0x32] = (vga->mode->vrs>>8) & 0x0F;
	vga->crt[0x33] = (vga->mode->vrs>>8) & 0x0F;
	vga->crt[0x41] = (vga->crt[0x13]>>8) & 0x0F;
	vga->crt[0x40] = 0x80;

	ct65550->xr[0x40] |= 0x03;
	ct65550->xr[0x80] |= 0x10;
	ct65550->xr[0x81] &= ~0x0F;
	if(vga->mode->z == 8)
		ct65550->xr[0x81] |= 0x12;

	ct65550->xr[0x0A] = 0x01;
	if(vga->linear && (ctlr->flag & Hlinear))
		ctlr->flag |= Ulinear;

	ctlr->flag |= Finit;
}

static void
load(Vga* vga, Ctlr* ctlr)
{
	Ct65550 *ct65550;

	ct65550 = vga->private;

	ct65550xo(Xrx, 0x0E, 0x00);
	while((vgai(Status1) & 0x08) == 0x08)
		;
	while((vgai(Status1) & 0x08) == 0)
		;
	vgaxo(Seqx, 0x07, 0x00);

	/*
	 * Set the clock if necessary.
	 */
	if((vga->misc & 0x0C) == 0x08 && !ct65550->biosclk){
		vgao(MiscW, vga->misc & ~0x0C);
		ct65550xo(Frx, 0x03, ct65550->fr[0x03] & ~0x0C);

		ct65550xo(Xrx, 0xC8, ct65550->xr[0xC8]);
		ct65550xo(Xrx, 0xC9, ct65550->xr[0xC9]);
		ct65550xo(Xrx, 0xCA, ct65550->xr[0xCA]);
		ct65550xo(Xrx, 0xCB, ct65550->xr[0xCB]);
	}
	ct65550xo(Frx, 0x03, ct65550->fr[0x03]);
	vgao(MiscW, vga->misc);

	ct65550xo(Xrx, 0x09, ct65550->xr[0x09]);
	vgaxo(Crtx, 0x30, vga->crt[0x30]);
	vgaxo(Crtx, 0x31, vga->crt[0x31]);
	vgaxo(Crtx, 0x32, vga->crt[0x32]);
	vgaxo(Crtx, 0x33, vga->crt[0x33]);
	vgaxo(Crtx, 0x41, vga->crt[0x41]);
	vgaxo(Crtx, 0x40, vga->crt[0x40]);

	ct65550xo(Xrx, 0x40, ct65550->xr[0x40]);
	ct65550xo(Xrx, 0x80, ct65550->xr[0x80]);
	ct65550xo(Xrx, 0x81, ct65550->xr[0x81]);
	if(ctlr->flag & Ulinear){
		ct65550->xr[0x05] = vga->vmb>>16;
		ct65550->xr[0x06] = vga->vmb>>24;
		ct65550xo(Xrx, 0x05, ct65550->xr[0x05]);
		ct65550xo(Xrx, 0x06, ct65550->xr[0x06]);
		ct65550->xr[0x0A] = 0x02;
	}
	ct65550xo(Xrx, 0x0A, ct65550->xr[0x0A]);

	ctlr->flag |= Fload;
}

static ulong
dumpmclk(uchar data[4])
{
	double f, m, n;
	long pds, rds;

	m = data[0] & 0x7F;
	n = data[1] & 0x7F;
	pds = 1<<((data[2] & 0x70)>>4);
	if(data[2] & 0x01)
		rds = 1;
	else
		rds = 4;
	f = (RefFreq*4*(m+2))/(rds*(n+2));
	f /= pds;

	return f;
}

static ulong
dumpvclk(uchar data[4])
{
	double f, m, n;
	long pds, rds;

	m = ((data[2] & 0x03)<<8)|data[0];
	n = (((data[2] & 0x30)>>4)<<8)|data[1];
	pds = 1<<((data[3] & 0x70)>>4);
	if(data[3] & 0x01)
		rds = 1;
	else
		rds = 4;
	f = (RefFreq*4*(m+2))/(rds*(n+2));
	f /= pds;

	return f;
}

static void
dump(Vga* vga, Ctlr* ctlr)
{
	long i;
	char *name;
	Ct65550 *ct65550;

	name = ctlr->name;
	ct65550 = vga->private;

	printitem(name, "Fr00");
	for(i = 0; i < 0x50; i++)
		printreg(ct65550->fr[i]);
	printitem(name, "Mr00");
	for(i = 0; i < 0x60; i++)
		printreg(ct65550->mr[i]);
	printitem(name, "Crt30");
	for(i = 0x30; i < 0x80; i++)
		printreg(vga->crt[i]);
	printitem(name, "Xr00");
	for(i = 0; i < 0x100; i++)
		printreg(ct65550->xr[i]);

	printitem(ctlr->name, "CLK0");
	for(i = 0; i < 4; i++)
		printreg(ct65550->xr[0xC0+i]);
	fprintf(stdout, "%23ld", dumpvclk(&ct65550->xr[0xC0]));
	printitem(ctlr->name, "CLK1");
	for(i = 0; i < 4; i++)
		printreg(ct65550->xr[0xC4+i]);
	fprintf(stdout, "%23ld", dumpvclk(&ct65550->xr[0xC4]));
	printitem(ctlr->name, "CLK2");
	for(i = 0; i < 4; i++)
		printreg(ct65550->xr[0xC8+i]);
	fprintf(stdout, "%23ld", dumpvclk(&ct65550->xr[0xC8]));
	printitem(ctlr->name, "MCLK");
	for(i = 0; i < 4; i++)
		printreg(ct65550->xr[0xCC+i]);
	fprintf(stdout, "%23ld", dumpmclk(&ct65550->xr[0xCC]));

	printitem(ctlr->name, "m n pd rd");
	fprintf(stdout, "%9ld %8ld       - %8ld %8ld\n",
		vga->m[0], vga->n[0], vga->p[0], vga->r[0]);

	printitem(ctlr->name, "biosclk");
	printreg(ct65550->biosclk);
}

Ctlr ct65550 = {
	"ct65550",			/* name */
	snarf,				/* snarf */
	options,			/* options */
	init,				/* init */
	load,				/* load */
	dump,				/* dump */
};

Ctlr ct65550hwgc = {
	"ct65550hwgc",			/* name */
	0,				/* snarf */
	0,				/* options */
	0,				/* init */
	0,				/* load */
	0,				/* dump */
};

Ctlr ct65550biosclk = {
	"ct65550biosclk",		/* name */
	0,				/* snarf */
	0,				/* options */
	biosclkinit,			/* init */
	0,				/* load */
	0,				/* dump */
};
