/*
 * Generic VGA registers.
 */
enum {
	MiscW		= 0x03C2,	/* Miscellaneous Output (W) */
	MiscR		= 0x03CC,	/* Miscellaneous Output (R) */
	Status0		= 0x03C2,	/* Input status 0 (R) */
	Status1		= 0x03DA,	/* Input Status 1 (R) */
	FeatureR	= 0x03CA,	/* Feature Control (R) */
	FeatureW	= 0x03DA,	/* Feature Control (W) */

	Seqx		= 0x03C4,	/* Sequencer Index, Data at Seqx+1 */
	Crtx		= 0x03D4,	/* CRT Controller Index, Data at Crtx+1 */
	Grx		= 0x03CE,	/* Graphics Controller Index, Data at Grx+1 */
	Attrx		= 0x03C0,	/* Attribute Controller Index and Data */

	PaddrW		= 0x03C8,	/* Palette Address Register, write */
	Pdata		= 0x03C9,	/* Palette Data Register */
	Pixmask		= 0x03C6,	/* Pixel Mask Register */
	PaddrR		= 0x03C7,	/* Palette Address Register, read */
	Pstatus		= 0x03C7,	/* DAC Status (RO) */

	Pcolours	= 256,		/* Palette */
	Red		= 0,
	Green		= 1,
	Blue		= 2,

	Pblack		= 0x00,
	Pwhite		= 0xFF,
};

#define RefFreq		14318180UL
#define VgaFreq0	25175000UL
#define VgaFreq1	28322000UL

typedef struct Ctlr Ctlr;
typedef struct Vga Vga;

typedef struct Ctlr {
	char	name[NAMELEN+1];
	void	(*snarf)(Vga *vganotused, Ctlr *ctlrnotused);
	void	(*options)(Vga *vganotused, Ctlr *ctlrnotused);
	void	(*init)(Vga *vganotused, Ctlr *ctlrnotused);
	void	(*load)(Vga *vganotused, Ctlr *ctlrnotused);
	void	(*dump)(Vga *vganotused, Ctlr *ctlrnotused);
	char*	type;

	ulong	flag;

	Ctlr*	link;
} Ctlr;

enum {					/* flag */
	Fsnarf		= 0x00000001,	/* snarf done */
	Foptions	= 0x00000002,	/* options done */
	Finit		= 0x00000004,	/* init done */
	Fload		= 0x00000008,	/* snarf done */
	Fdump		= 0x00000010,	/* dump done */

	Hpclk2x8	= 0x00000100,	/* have double 8-bit mode */
	Upclk2x8	= 0x00000200,	/* use double 8-bit mode */
	Henhanced	= 0x00000400,	/* have enhanced mode */
	Uenhanced	= 0x00000800,	/* use enhanced mode */
	Hpvram		= 0x00001000,	/* have parallel VRAM */
	Upvram		= 0x00002000,	/* use parallel VRAM */
	Hextsid		= 0x00004000,	/* have external SID mode */
	Uextsid		= 0x00008000,	/* use external SID mode */

};
/* too big for 16-bit enum */

#define	Hclk2		 0x00010000UL	/* have clock-doubler */
#define	Uclk2		 0x00020000UL	/* use clock-doubler */
#define	Hlinear		 0x00040000UL	/* have linear-address mode */
#define	Ulinear		 0x00080000UL	/* use linear-address mode */
#define	Hclkdiv		 0x00100000UL	/* have a clock-divisor */
#define	Uclkdiv		 0x00200000UL	/* use clock-divisor */
#define	Hsid32		 0x00400000UL	/* have a 32-bit (as opposed to 64-bit) SID */


typedef struct Attr Attr;
typedef struct Attr {
	char*	attr;
	char*	val;

	Attr*	next;
} Attr;

typedef struct Mode {
	char	type[NAMELEN+1];	/* monitor type e.g. "vs1782" */
	char	size[NAMELEN+1];	/* size e.g. "1376x1024x8" */
	char chan[NAMELEN+1];	/* channel descriptor, e.g. "m8" or "r8g8b8a8" */

	long	frequency;		/* Dot Clock (MHz) */
	long	deffrequency;		/* Default dot clock if calculation can't be done */
	long	x;			/* Horizontal Display End (Crt01), from .size[] */
	long	y;			/* Vertical Display End (Crt18), from .size[] */
	long	z;			/* depth, from .size[] */

	long	ht;			/* Horizontal Total (Crt00) */
	long	shb;			/* Start Horizontal Blank (Crt02) */
	long	ehb;			/* End Horizontal Blank (Crt03) */

	long	shs;			/* optional Start Horizontal Sync (Crt04) */
	long	ehs;			/* optional End Horizontal Sync (Crt05) */

	long	vt;			/* Vertical Total (Crt06) */
	long	vrs;			/* Vertical Retrace Start (Crt10) */
	long	vre;			/* Vertical Retrace End (Crt11) */

	ulong	videobw;

	char	hsync;
	char	vsync;
	char	interlace;

	Attr*	attr;
} Mode;

/*
 * The sizes of the register sets are large as many SVGA and GUI chips have extras.
 * The Crt registers are ushorts in order to keep overflow bits handy.
 * The clock elements are used for communication between the VGA, RAMDAC and clock chips;
 * they can use them however they like, it's assumed they will be used compatibly.
 */
typedef struct Vga {
	uchar	misc;
	uchar	feature;
	uchar	sequencer[256];
	ushort	crt[256];
	uchar	graphics[256];
	uchar	attribute[256];
	uchar	pixmask;
	uchar	pstatus;
	uchar	palette[Pcolours][3];

	ulong	f[2];			/* clock */
	ulong	d[2];
	ulong	i[2];
	ulong	m[2];
	ulong	n[2];
	ulong	p[2];
	ulong	q[2];
	ulong	r[2];

	ulong	vma;			/* video memory linear-address alignment */
	ulong	vmb;			/* video memory linear-address base */
	ulong	vmz;			/* video memory size */

	ulong	membw;		/* memory bandwidth, MB/s */

	long	offset;			/* BIOS string offset */
	char*	bios;			/* matching BIOS string */

	Mode*	mode;

	Ctlr*	ctlr;
	Ctlr*	ramdac;
	Ctlr*	clock;
	Ctlr*	hwgc;
	Ctlr*	link;
	long	linear;
	Attr*	attr;

	void*	private;
} Vga;

/* ark2000pv.c */
extern Ctlr ark2000pv;
extern Ctlr ark2000pvhwgc;

/* att20c49x.c */
extern Ctlr att20c490;
extern Ctlr att20c491;
extern Ctlr att20c492;

/* att21c498.c */
extern uchar attdaci(uchar);
extern void attdaco(uchar, uchar);
extern Ctlr att21c498;

/* bt485.c */
extern uchar bt485i(uchar);
extern void bt485o(uchar, uchar);
extern Ctlr bt485;

/* ch9294.c */
extern Ctlr ch9294;

/* clgd542x.c */
extern Ctlr clgd542x;
extern Ctlr clgd542xhwgc;

/* ct65540.c */
extern Ctlr ct65540;
extern Ctlr ct65545;
extern Ctlr ct65545hwgc;

/* ct65550.c */
extern Ctlr ct65550;
extern Ctlr ct65550hwgc;
extern Ctlr ct65550biosclk;

/* cyber938x.c */
extern Ctlr cyber938x;
extern Ctlr cyber938xhwgc;

/* data.c */
extern long cflag;
extern long dflag;
extern Ctlr *ctlrs[];
extern ushort dacxreg[4];

/* db.c */
extern char* dbattr(Attr*, char*);
extern long dbctlr(char*, Vga*);
extern Mode* dbmode(char*, char*, char*);
extern void dbdumpmode(Mode*);

/* error.c */
extern void error(char*, ...);
extern void trace(char*, ...);
extern long vflag, Vflag;

/* et4000.c */
extern Ctlr et4000;

/* et4000hwgc.c */
extern Ctlr et4000hwgc;

/* ibm8514.c */
extern Ctlr ibm8514;

/* icd2061a.c */
extern Ctlr icd2061a;

/* ics2494.c */
extern Ctlr ics2494;
extern Ctlr ics2494a;

/* ics534x.c */
extern Ctlr ics534x;

/* io.c */
#undef inportb
#undef outportb
#undef inportw
#undef outportw

#define inportb _p9_inportb
#define outportb _p9_outportb
#define inportw _p9_inportw
#define outportw _p9_outportw

extern uchar inportb(long);
extern void outportb(long, uchar);
extern ushort inportw(long);
extern void outportw(long, ushort);
extern ulong inportl(long);
extern void outportl(long, ulong);
extern char* vgactlr(char*, char*);
extern void vgactlw(char*, char*);
extern long readbios(char*, long, long);
extern void dumpbios(void);
extern void error(char*, ...);
extern void* alloc(ulong);
extern void printitem(char*, char*);
extern void printreg(ulong);
extern void printflag(ulong);
extern void setpalette(long, long, long, long);
extern long curprintindex;

/* mach32.c */
extern Ctlr mach32;

/* mach64.c */
extern Ctlr mach64;

/* mach64xx.c */
extern Ctlr mach64xx;
extern Ctlr mach64xxhwgc;

/* main.c */
extern char* chanstr[];
extern char* ramdacstr;
extern char* linkstr;
extern char* clockstr;
extern char* hwgcstr;
extern void resyncinit(Vga *vganotused, Ctlr*, ulong, ulong);
extern void sequencer(Vga *vganotused, long);


/* mga2164w.c */
extern Ctlr mga2164w;
extern Ctlr mga2164whwgc;

/* nvidia.c */
extern Ctlr nvidia;
extern Ctlr nvidiahwgc;

/* palette.c */
extern Ctlr palette;

/* pci.c */
typedef struct Pcidev Pcidev;

/* this is here rather than pci.h to shut bcc up. */
struct Pcidev {
	long	tbdf;			/* type+bus+device+function */
	ushort	vid;			/* vendor ID */
	ushort	did;			/* device ID */

	struct {
		ulong	bar;		/* base address */
		long	size;
	} mem[3];

	uchar	intl;			/* interrupt line */
	ushort	ccru;


	Pcidev*	list;
	Pcidev*	bridge;			/* down a bus */
	Pcidev*	link;			/* next device on this bno */
};

extern long pcicfgr8(Pcidev*, long);
extern long pcicfgr16(Pcidev*, long);
extern long pcicfgr32(Pcidev*, long);
extern void pcicfgw8(Pcidev*, long, long);
extern void pcicfgw16(Pcidev*, long, long);
extern void pcicfgw32(Pcidev*, long, long);
extern void pcihinv(Pcidev*);
extern Pcidev* pcimatch(Pcidev*, long, long);

/* rgb524.c */
extern Ctlr rgb524;

/* s3801.c */
extern Ctlr s3801;
extern Ctlr s3805;

/* s3928.c */
extern Ctlr s3928;

/* s3clock.c */
extern Ctlr s3clock;

/* s3generic.c */
extern Ctlr s3generic;

/* s3hwgc.c */
extern Ctlr bt485hwgc;
extern Ctlr rgb524hwgc;
extern Ctlr s3hwgc;
extern Ctlr tvp3020hwgc;
extern Ctlr tvp3026hwgc;

/* sc15025.c */
extern Ctlr sc15025;

/* stg1702.c */
extern Ctlr stg1702;

/* trio64.c */
extern void trio64clock(Vga *vganotused, Ctlr *ctlrnotused);
extern Ctlr trio64;

/* tvp3020.c */
extern uchar tvp3020i(uchar);
extern uchar tvp3020xi(uchar);
extern void tvp3020o(uchar, uchar);
extern void tvp3020xo(uchar, uchar);
extern Ctlr tvp3020;

/* tvp3025.c */
extern Ctlr tvp3025;

/* tvp3025clock.c */
extern Ctlr tvp3025clock;

/* tvp3026.c */
extern uchar tvp3026xi(uchar);
extern void tvp3026xo(uchar, uchar);
extern Ctlr tvp3026;

/* tvp3026clock.c */
extern Ctlr tvp3026clock;

/* vga.c */
extern uchar vgai(long);
extern uchar vgaxi(long, uchar);
extern void vgao(long, uchar);
extern void vgaxo(long, uchar, uchar);
extern Ctlr generic;

/* virge.c */
extern Ctlr virge;

/* vision864.c */
extern Ctlr vision864;

/* vision964.c */
extern Ctlr vision964;

/* vision968.c */
extern Ctlr vision968;

/* w30c516.c */
extern Ctlr w30c516;
